
package org.saev.deob.graph.hier;

import java.util.*;
import org.apache.regexp.*;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.saev.deob.Application;
import org.saev.deob.graph.*;
import org.saev.deob.util.*;


public class ClassHierarchy extends HashMutableDirectedGraph<JavaClass> {
    private Application app;
    
    private static final String INIT = "<init>";
    private static final String CINIT = "<clinit>";
    
    private static final String[] LIB = { "sign[.]", "java[.].*", "javax[.].*",
            "sun[.].*", "sunw[.].*", "com[.]sun[.].*", "org[.]omg[.].*",
            "org[.]w3c[.].*", "org[.]xml[.].*", "net[.]jini[.].*"
    };
    
    public ClassHierarchy(Application app) {
        this.app = app;
        for(ClassGen cgen: app.getClasses()) {
            addClass(cgen.getJavaClass());
        }
    }
    
    public void addClass(JavaClass clazz) {
        
        if(containsNode(clazz)) {
            return;
        }
        addNode(clazz);
        
        try {
            JavaClass superClass = clazz.getSuperClass();
            
            if (superClass != null) {
                addClass(superClass);
                addEdge(superClass, clazz);
            }
            
            JavaClass[] interfaces = clazz.getInterfaces();
            
            for(int i = 0; i < interfaces.length; i++) {
                addClass(interfaces[i]);
                addEdge(interfaces[i], clazz);
            }
        } catch(ClassNotFoundException e) {
            throw new RuntimeException(e);
        }
    }
    
    public JavaClass[] superClasses(JavaClass subClass) {
        return (JavaClass[]) getPreds(subClass).toArray(new JavaClass[0]);
    }
    
    public JavaClass[] subClasses(JavaClass superClass) {
        return (JavaClass[]) getSuccs(superClass).toArray(new JavaClass[0]);
    }
    
    public JavaClass[] inheritanceChain(JavaClass clazz) {
        JavaClass[] sub = subClasses(clazz);
        JavaClass[] sup = superClasses(clazz);
        
        JavaClass[] all = new JavaClass[sub.length + sup.length + 1];
        all[0] = clazz;
        int idx = 1;
        
        for(int i = 0 ; i < sub.length ; i++) {
            
            if(sub[i] != clazz) {
                all[idx++] = sub[i];
            }
        }
        
        for(int i = 0 ; i < sup.length ; i++) {
            
            if(sup[i] != clazz) {
                all[idx++] = sup[i];
            }
        }
        return all;
    }
    
    public boolean classExtends(JavaClass subClass, JavaClass superClass) {
        return getPreds(subClass).contains(superClass) ||
                getSuccs(subClass).contains(superClass);
    }
    
    public boolean overrides(MethodID subMethod, MethodID superMethod) {
        JavaClass subClass = app.getJavaClass(subMethod.getClassName());
        JavaClass superClass = app.getJavaClass(superMethod.getClassName());
        
        if (!subMethod.getName().equals(superMethod.getName())) {
            return false;
        }
        
        if (!subMethod.getSignature().equals(superMethod.getSignature())) {
            return false;
        }
        return classExtends(subClass, superClass);
    }
    
    public boolean overloads(MethodID subMethod, MethodID superMethod) {
        JavaClass subClass = app.getJavaClass(subMethod.getClassName());
        JavaClass superClass = app.getJavaClass(superMethod.getClassName());
        
        if (!subMethod.getName().equals(superMethod.getName())) {
            return false;
        }
        
        if (subMethod.getSignature().equals(superMethod.getSignature())) {
            return false;
        }
        return classExtends(subClass, superClass);
    }
    
    public MethodID[] overrides(MethodID subMethod) {
        JavaClass subClass = app.getJavaClass(subMethod.getClassName());
        ArrayList<MethodID> res = new ArrayList<MethodID>();
        JavaClass[] superClasses = superClasses(subClass);
        
        for(int i = 0; i < superClasses.length; i++) {
            JavaClass curr = superClasses[i];
            Method[] methods = curr.getMethods();
            
            for(int j = 0; j < methods.length; j++) {
                Method m = methods[j];
                MethodID mid = new MethodID(m.getName(), m.getSignature(), curr.getClassName());
                
                if (overrides(subMethod, mid)) {
                    res.add(mid);
                }
            }
        }
        return res.toArray(new MethodID[0]);
    }
    
    public MethodID[] allOverrides(MethodID method) {
        ArrayList<MethodID> res = new ArrayList<MethodID>();
        JavaClass[] classes = inheritanceChain(app.getJavaClass(method.getClassName()));
        
        for(int i = 0; i < classes.length; i++) {
            JavaClass curr = classes[i];
            Method[] methods = curr.getMethods();
            
            for(int j = 0; j < methods.length; j++) {
                Method m = methods[j];
                MethodID mid = new MethodID(m.getName(), m.getSignature(), curr.getClassName());
                
                if (method.getName().equals(mid.getName()) &&
                        method.getSignature().equals(mid.getSignature())) {
                    res.add(mid);
                }
            }
        }
        return res.toArray(new MethodID[0]);
    }
    
    public MethodID[] allOverloads(MethodID method) {
        ArrayList<MethodID> res = new ArrayList<MethodID>();
        JavaClass[] classes = inheritanceChain(app.getJavaClass(method.getClassName()));
        
        for(int i = 0; i < classes.length; i++) {
            JavaClass curr = classes[i];
            Method[] methods = curr.getMethods();
            
            for(int j = 0; j < methods.length; j++) {
                Method m = methods[j];
                MethodID mid = new MethodID(m.getName(), m.getSignature(), curr.getClassName());
                
                if (method.getName().equals(mid.getName()) &&
                        !method.getSignature().equals(mid.getSignature())) {
                    res.add(mid);
                }
            }
        }
        return res.toArray(new MethodID[0]);
    }
    
    public boolean methodRenameOK(MethodID method, String newName) {
        MethodID newMethod = new MethodID(newName, method.getSignature(), method.getClassName());
        return methodRenameOK(method, newMethod);
    }
    
    public boolean methodRenameOK(MethodID method, MethodID newMethod) {
        
        if(isSpecialMethod(method)) {
            return false;
        }
        
        if(overridesLibraryMethod(method)) {
            return false;
        }
        MethodID[] overriddenNewMethods = overrides(newMethod);
        
        if (overriddenNewMethods.length > 0) {
            return false;
        }
        return true;
    }
    
    public boolean isLibraryClass(String className) {
        
        try {
            
            for(int i = 0; i < LIB.length; i++) {
                RE regex = new RE(LIB[i]);
                
                if(regex.match(className)) {
                    return true;
                }
            }
        } catch(RESyntaxException e) {
            throw new RuntimeException(e);
        }
        return false;
    }
    
    public boolean overridesLibraryMethod(MethodID method) {
        MethodID[] overiddenMethods = overrides(method);
        for(int i = 0; i < overiddenMethods.length; i++) {
            JavaClass clazz = app.getJavaClass(overiddenMethods[i].getClassName());
            
            if(clazz != null && isLibraryClass(clazz.getClassName())) {
                return true;
            }
        }
        return false;
    }
    
    public boolean isSpecialMethod(MethodID method) {
        
        if(method.getName().equals(CINIT) || method.getName().equals(INIT)){
            return true;
        }
        
        if(method.getName().equals("main") && method.getSignature().equals("([Ljava/lang/String;)V")) {
            return true;
        }
        return false;
    }
    
    public boolean isInitMethod(Method method) {
        return (method.getName().equals(INIT) || method.getName().equals(CINIT));
    }
    
    public MethodGen resolveMethodReference(MethodID mid, JavaClass referent) throws ClassNotFoundException {
        JavaClass container = app.getClassGen(mid.getClassName()).getJavaClass();
        
        if (container == null || container.isInterface()) {
            return null;
        }
        MethodGen method = null;
	    JavaClass superClass = container;
            
		while(method == null) {
			method = app.getMethod(superClass.getClassName(), mid.getName(), mid.getSignature());
                
			if(superClass.getSuperClass() == superClass) {
				break;
			}
			superClass = superClass.getSuperClass();
		}
        
        if (method == null) {
            HashSet<JavaClass> visited = new HashSet<JavaClass>();
            LinkedList<JavaClass> queue = new LinkedList<JavaClass>();
            queue.addAll(Arrays.asList(container.getInterfaces()));
            
            while(method == null && !queue.isEmpty()) {
                JavaClass iface = queue.removeFirst();
                
                if(visited.contains(iface)) {
                    continue;
                }
                visited.add(iface);
                method = app.getMethod(iface.getClassName(), mid.getName(), mid.getSignature());
                queue.addAll(Arrays.asList(iface.getInterfaces()));
            }
        }
        
        if (method == null || (method.isAbstract() && !container.isAbstract()) ||
                !isAccessible(new MethodID(method), referent)) {
            return null;
        } else
            return method;
    }
    
    public MethodGen resolveInterfaceMethodReference(MethodID mid, JavaClass referent) throws ClassNotFoundException {
        JavaClass container = app.getClassGen(mid.getClassName()).getJavaClass();;
        
        if(container == null || !container.isInterface()) {
            return null;
        }
        MethodGen method = app.getMethod(mid);
        
        if (method == null) {
            HashSet<JavaClass> visited = new HashSet<JavaClass>();
            LinkedList<JavaClass> queue = new LinkedList<JavaClass>();
            queue.addAll(Arrays.asList(container.getInterfaces()));
            
            while(method == null && !queue.isEmpty()) {
                JavaClass iface = queue.removeFirst();
                
                if(visited.contains(iface)) {
                    continue;
                }
                visited.add(iface);
                method = app.getMethod(iface.getClassName(), mid.getName(), mid.getSignature());
                queue.addAll(Arrays.asList(iface.getInterfaces()));
            }
        }
        
        if(method == null) {
            JavaClass superClass = container.getSuperClass();
            
            if(!superClass.getClassName().equals("java.lang.Object")) {
                throw new RuntimeException("interfaces should have object as superclass: " + superClass.getClassName());
            }
            method = app.getMethod(superClass.getClassName(), mid.getName(), mid.getSignature());
        }
        return method;
    }
    
    public FieldGen resolveFieldReference(FieldID fid, JavaClass referent) throws ClassNotFoundException {
        return resolveFieldReference(fid, app.getClassGen(fid.getClassName()).getJavaClass(), referent);
    }
    
    public FieldGen resolveFieldReference(FieldID fid, JavaClass container, JavaClass referent) throws ClassNotFoundException {
        
        if(container == null) {
            return null;
        }
        FieldGen field = app.getField(container.getClassName(), fid.getName(), fid.getSignature());
        
        if(field == null) {
            JavaClass interfaces[] = container.getInterfaces();
            
            for(int i = 0; field == null && i < interfaces.length; i++) {
                field = resolveFieldReference(fid, interfaces[i], referent);
            }
        }
        if(field == null && container.getSuperClass() != container) {
            field = resolveFieldReference(fid, container.getSuperClass(), referent);
        }
        
        if(field == null || !isAccessible(fid, referent)) {
            return null;
        }
        return field;
    }
    
    public boolean isAccessible(MethodID mid, JavaClass clazz) {
        MethodGen mg = app.getMethod(mid);
        
        if(mg != null) {
            
            if(mg.isPublic()) {
                return true;
            }
            JavaClass container = app.getJavaClass(mid.getClassName());
            
            if(mg.isProtected() && classExtends(clazz, container)) {
                return true;
            }
            
            if((mg.isProtected() || (!mg.isPublic() && !mg.isProtected() && !mg.isPrivate())) &&
                    clazz.getPackageName().equals(container.getPackageName())) {
                return true;
            }
            
            if(mg.isPrivate() && container == clazz) {
                return true;
            }
        }
        return false;
    }
    
    public boolean isAccessible(FieldID fid, JavaClass clazz) {
        FieldGen fg = app.getField(fid);
        
        if(fg != null) {
            
            if(fg.isPublic()) {
                return true;
            }
            JavaClass container = app.getJavaClass(fid.getClassName());
            
            if(fg.isProtected() && classExtends(clazz, container)) {
                return true;
            }
            
            if((fg.isProtected() || (!fg.isPublic() && !fg.isProtected() && !fg.isPrivate())) &&
                    clazz.getPackageName().equals(container.getPackageName())) {
                return true;
            }
            
            if(fg.isPrivate() && container == clazz) {
                return true;
            }
            
            if (fg.isPrivate() && app.getMethod(fid.getClassName(), fid.getName(), fid.getSignature()) != null) {
                return true;
            }
        }
        return false;
    }
    
    public String toString() {
        StringBuilder sb = new StringBuilder("");
        Iterator it = iterator();
        
        while(it.hasNext()) {
            JavaClass clazz = (JavaClass) it.next();
            sb.append(clazz.getClassName());
            JavaClass[] classes = inheritanceChain(clazz);
            
            for(int i = 1; i < classes.length; i++) {
                sb.append(" -> " + classes[i].getClassName());
            }
            sb.append("\n");
        }
        return sb.toString();
    }
}